local addonName, addon = ...

-- functions
local ipairs, pairs, type, unpack, wipe = ipairs, pairs, type, unpack, wipe
local format = string.format
local tconcat, tinsert, tsort = table.concat, table.insert, table.sort
local ChatEdit_GetActiveWindow = ChatEdit_GetActiveWindow
local ChatEdit_InsertLink = ChatEdit_InsertLink
local ChatFrame_OpenChat = ChatFrame_OpenChat
local CreateFrame = CreateFrame
local EasyMenu = EasyMenu
local GetAutoCompleteRealms = GetAutoCompleteRealms
local IsInGroup = IsInGroup
local IsInGuild = IsInGuild
local IsShiftKeyDown = IsShiftKeyDown
local SendChatMessage = SendChatMessage
-- constants
local DEFAULT_CHAT_FRAME = DEFAULT_CHAT_FRAME
local LE_PARTY_CATEGORY_INSTANCE = LE_PARTY_CATEGORY_INSTANCE
-- frames
local GameTooltip = GameTooltip
local UIParent = UIParent
-- libs
local LibStub = LibStub

local broker = LibStub:GetLibrary("LibDataBroker-1.1"):NewDataObject(addonName, {
    type = "data source",
    label = "Keystone",
    text = "No Keystone",
    icon = [[Interface\Icons\inv_relics_hourglass]],
})
broker.menuFrame = CreateFrame("frame", "BrokerKeystonesContextMenuFrame", UIParent, "UIDropDownMenuTemplate")

function broker:GetTooltipLines()
    local lines = {}

    local charactersWithKeystones = addon:GetCharacters(function(_, c) return c.keystone ~= nil end)
    local numKeystones = addon:TableLength(charactersWithKeystones)

    local charactersWithRecords = addon.options.showRecords
        and addon:GetCharacters(function(_, c) return c.record ~= nil end)
        or {}
    local numRecords = addon:TableLength(charactersWithRecords)

    if numKeystones < 1 and numRecords < 1 then
        tinsert(lines, {"No Keystones", nil, 1, 1, 1})
        return lines
    end

    tinsert(lines, {
        "Character",
        format("%d |cffFFFFFF%s|r", numKeystones, numKeystones == 1 and "Keystone" or "Keystones"),
        1, 1, 1,
    })
    if not addon.options.showRecords then tinsert(lines, " ") end

    -- merge the two tables
    local characters = {}
    if numKeystones > 0 then
        for k, v in pairs(charactersWithKeystones) do
            characters[k] = v
        end
    end
    if numRecords > 0 then
        for k, v in pairs(charactersWithRecords) do
            characters[k] = v
        end
    end

    for key, character in pairs(characters) do
        if addon.options.showRecords then tinsert(lines, " ") end
        tinsert(lines, {
            addon:GetCharacterDisplayName(key, true),
            character.keystone
            and addon:GetKeystoneDisplayText(character.keystone, true)
            or format("|cff666666%s|r", "(no keystone)")
        })
        if addon.options.showRecords and character.record then
            local mapId, level = unpack(character.record)
            -- todo: shorthand
            local instanceName = addon:GetMapName(mapId)
            tinsert(lines, {
                " ",
                format("Best: %s %d", instanceName, level),
                nil, nil, nil,
                0.7, 0.7, 0.7,
            })
        end
    end

    return lines
end

function broker:OnTooltipShow()
    local lines = broker:GetTooltipLines()

    for _, line in pairs(lines) do
        if type(line) == "string" then
            line = {line}
        end

        self:AddDoubleLine(unpack(line))
    end
end

function broker:OnClick(btn)
    if btn == "LeftButton" then return broker:OnLeftClick() end
    if btn == "RightButton" then return broker:OnRightClick() end
end

function broker:ContextMenu(menu)
    EasyMenu(menu, broker.menuFrame, "cursor", 0, 0, "MENU")
end

local t = {}
local function BuildKeystoneList(callback)
    wipe(t)
    local realms = {}
    for _, realm in ipairs(GetAutoCompleteRealms()) do
        realms[realm] = true
    end
    for key, character in addon:GetCharactersIterator(function(_, c)
        return c.keystone ~= nil and c.faction == addon.player.faction and realms[c.realm]
    end) do
        tinsert(t, format("%s (%s)",
            addon:GetKeystoneDisplayText(character.keystone),
            addon:GetCharacterClassDisplayName(key)
        ))
    end

    if #t > 0 then
        tsort(t)
        return callback(nil, tconcat(t, ", "))
    end

    return callback("No keystones to report.")
end

local function BuildRecordList(callback)
    wipe(t)
    local realms = {}
    for _, realm in ipairs(GetAutoCompleteRealms()) do
        realms[realm] = true
    end
    for key, character in addon:GetCharactersIterator(function(_, c)
        return c.record ~= nil and c.faction == addon.player.faction and realms[c.realm]
    end) do
        local mapId, level = unpack(character.record)
        tinsert(t, format("%s %d (%s)",
            addon:GetMapName(mapId),
            level,
            addon:GetCharacterClassDisplayName(key)
        ))
    end

    if #t > 0 then
        tsort(t)
        return callback(nil, tconcat(t, ", "))
    end

    return callback("No records to report.")
end

function broker:AddLinkToChat(link)
    local chatWindow = ChatEdit_GetActiveWindow()
    if chatWindow then
        ChatEdit_InsertLink(link)
    else
        ChatFrame_OpenChat(link, DEFAULT_CHAT_FRAME)
    end
end

function broker:AddTextToChat(text)
    local chatWindow = ChatEdit_GetActiveWindow()
    if chatWindow then
        chatWindow:Insert(text)
    else
        ChatFrame_OpenChat(text, DEFAULT_CHAT_FRAME)
    end
end

function broker:OnLeftClick()
    if IsShiftKeyDown() then
        if addon.player.keystone == nil then
            addon:Print("You don't have a keystone on this character.")
            addon:Print("If you did, shift-clicking the broker would link it to your chat.")

            return
        end

        return broker:AddLinkToChat(addon.player.keystone)
    end

    broker:ContextMenu({
        {text = "Report Keystones in Chat Box", notCheckable = true, func = function()
            BuildKeystoneList(function(err, list)
                if err then addon:Print(err) return end

                broker:AddTextToChat(format("Keystones: %s", list))
            end)
        end},
        {text = "Report Keystones to Guild", notCheckable = true, func = function()
            BuildKeystoneList(function(err, list)
                if err then addon:Print(err) return end

                if IsInGuild() then
                    SendChatMessage(format("Keystones: %s", list), "GUILD")
                else
                    addon:Print("You are not in a guild.")
                end
            end)
        end},
        {text = "Report Keystones to Group", notCheckable = true, func = function()
            BuildKeystoneList(function(err, list)
                if err then addon:Print(err) return end

                if IsInGroup(LE_PARTY_CATEGORY_INSTANCE) and not IsShiftKeyDown() then
                    SendChatMessage(format("Keystones: %s", list), "INSTANCE_CHAT")
                elseif IsInGroup() then
                    SendChatMessage(format("Keystones: %s", list), "PARTY")
                else
                    addon:Print("You are not in a party or instance.")
                end
            end)
        end},
        {text = "- - - - - - - - - -", disabled = true, notCheckable = true},
        {text = "Report Weekly Records in Chat Box", notCheckable = true, func = function()
            BuildRecordList(function(err, list)
                if err then addon:Print(err) return end

                broker:AddTextToChat(format("Weekly Records: %s", list))
            end)
        end},
        {text = "Report Weekly Records to Guild", notCheckable = true, func = function()
            BuildRecordList(function(err, list)
                if err then addon:Print(err) return end

                if IsInGuild() then
                    SendChatMessage(format("Weekly Records: %s", list), "GUILD")
                else
                    addon:Print("You are not in a guild.")
                end
            end)
        end},
        {text = "Report Weekly Records to Group", notCheckable = true, func = function()
            BuildRecordList(function(err, list)
                if err then addon:Print(err) return end

                if IsInGroup(LE_PARTY_CATEGORY_INSTANCE) and not IsShiftKeyDown() then
                    SendChatMessage(format("Weekly Records: %s", list), "INSTANCE_CHAT")
                elseif IsInGroup() then
                    SendChatMessage(format("Weekly Records: %s", list), "PARTY")
                else
                    addon:Print("You are not in a party or instance.")
                end
            end)
        end},
    })
end

function broker:OnRightClick()
    broker:ContextMenu({
        {text = addonName, isTitle = true, notCheckable = true},
        {text = "Options", notCheckable = true, hasArrow = true, menuList = {
            {text = "Use shorthand dungeon names where available",
                keepShownOnClick = true,
                isNotRadio = true,
                checked = function() return addon.options.useShorthand end,
                func = function() addon.options.useShorthand = not addon.options.useShorthand end},
            {text = "Show characters' weekly records in tooltip",
                keepShownOnClick = true,
                isNotRadio = true,
                checked = function() return addon.options.showRecords end,
                func = function() addon.options.showRecords = not addon.options.showRecords end},
        }},
        {text = "Miscellaneous", notCheckable = true, hasArrow = true, menuList = {
            {text = "Delete Keystone and Record Data (no confirmation)",
                notCheckable = true,
                func = function()
                    addon:DeleteAllCharacterKeystones()
                    addon:DeleteAllCharacterRecords()
                end},
            {text = "Debug mode",
                keepShownOnClick = true,
                isNotRadio = true,
                checked = function() return addon.options.debug end,
                func = function() addon.options.debug = not addon.options.debug
            end},
        }},
    })
end

addon:RegisterCallback(addon.EVENT_KEYSTONE, function(_, key, link)
    broker.text = link and addon:GetKeystoneDisplayText(link, true) or "No Keystone"
end)

addon:RegisterCallback(addon.EVENT_OPTION_CHANGED, function(_, key)
    if key ~= "useShorthand" then return end

    local character = addon:GetCharacter()
    if not character or not character.keystone then
        broker.text = "No Keystone"
        return
    end

    broker.text = addon:GetKeystoneDisplayText(character.keystone, true)
end)
