local _, addon = ...

-- functions
local pairs, setmetatable = pairs, setmetatable
local format = string.format
local FillLocalizedClassList = FillLocalizedClassList
local GetRealmName = GetRealmName
local UnitClass = UnitClass
local UnitFactionGroup = UnitFactionGroup
local UnitName = UnitName
-- constants
local RAID_CLASS_COLORS = RAID_CLASS_COLORS
-- frames
-- libs

function addon:GetPlayerInfo()
    return {
        realm = GetRealmName(),
        faction = UnitFactionGroup("player"),
        name = select(1, UnitName("player")),
        class = select(2, UnitClass("player")),
    }
end

function addon:GetCharacterKey(character)
    if character then
        return format("%s.%s.%s", character.realm, character.faction, character.name)
    end

    local realm = GetRealmName()
    local name = select(1, UnitName("player"))
    local faction = UnitFactionGroup("player")

    return format("%s.%s.%s", realm, faction, name)
end

function addon:GetCharacter(key)
    key = key or self:GetCharacterKey()
    if not self.db.characters[key] then
        addon:Debug(format("Error: could not access character `%s`.", key or "(nil key)"))
        return
    end

    return self.db.characters[key]
end

function addon:GetCharacterKeystone(key)
    local character = self:GetCharacter(key)
    if not character then return end

    return character and character.keystone
end

function addon:SetCharacterKeystone(key, link)
    local character = self:GetCharacter(key)
    if not character then return end

    -- TODO: This checks if the character once had a keystone, and now does not.
    -- It will do terrible things if the player manually deletes their keystone.
    -- This should be handled better, but right now...
    if character.keystone ~= nil and link == nil then
        self:DeleteAllCharacterKeystones()
        self:DeleteAllCharacterRecords()
    end

    if link ~= character.keystone then
        character.keystone = link
    end

    self:Fire(self.EVENT_KEYSTONE, key, link)
end

function addon:SetCharacterRecord(key, mapId, level)
    local character = self:GetCharacter(key)
    if not character then return end

    if mapId == nil then
        character.record = nil
    else
        character.record = {mapId, level}
    end

    self:Fire(self.EVENT_RECORD, key, mapId, level)
end

function addon:GetCharacters(filter)
    if not filter then
        return self.db.characters
    end

    local len = 0
    local t = {}
    for k, v in pairs(self.db.characters) do
        if filter(k, v) then
            len = len + 1
            t[k] = v
        end
    end

    return setmetatable(t, {
        __len = function() return len end,
    })
end

function addon:TableLength(t)
    local i = 0

    for x, y in pairs(t) do
        i = i + 1
    end

    return i
end

function addon:GetCharactersIterator(filter)
    return pairs(self:GetCharacters(filter))
end

function addon:DeleteAllCharacterKeystones()
    for _, character in self:GetCharactersIterator() do
        character.keystone = nil
    end

    self:Fire(self.EVENT_KEYSTONE, self:GetCharacterKey(), nil)
    self:BagScannerOnEvent()
end

function addon:DeleteAllCharacterRecords()
    for _, character in self:GetCharactersIterator() do
        character.record = nil
    end

    self:Fire(self.EVENT_RECORD, self:GetCharacterKey())
end



function addon:GetCharacterName(key, colour)
    local character = addon:GetCharacter(key)
    if not character then return end

    if character.realm ~= self.player.realm then
        if colour then
            return format("|c%s%s-%s|r", RAID_CLASS_COLORS[character.class].colorStr, character.name, character.realm)
        end

        return format("%s-%s", character.name, character.realm)
    end

    if colour then
        return format("|c%s%s|r", RAID_CLASS_COLORS[character.class].colorStr, character.name)
    end

    return character.name
end

function addon:GetFactionName(faction, colour)
    if colour then
        local hex = "FF0000"
        if faction == "Alliance" then
            hex = "2459FF"
        end

        return format("|cff%s%s|r", hex, faction)
    end

    return faction
end

function addon:GetCharacterDisplayName(key, colour)
    local character = self:GetCharacter(key)
    if not character then return end

    if character.faction ~= self.player.faction then
        return format("%s (%s)", self:GetCharacterName(key, colour), self:GetFactionName(character.faction, colour))
    end

    return self:GetCharacterName(key, colour)
end

-- todo: cleanup
local _classNames = {}
FillLocalizedClassList(_classNames)
function addon:GetCharacterClassDisplayName(key, colour)
    local character = self:GetCharacter(key)
    if not character then return end

    return _classNames[character.class]
end
